/* *******************************************************************
Purpose:Compute the periodogram by naive implementation; smooth on
        log-periodogram; and local Whittle likelihood estimation of
	the spectral density
Usage:  To compile, type "cc autovar.c -lm".
        To run, type "a.out" and then follow the instructions.
Input:  n = sample size, length of the time series
        ngrid = number of grid point.  The current program use
		ngrid=201.  This can be changed by hand.
        hgiven = given bandwidth; if hgiven < 0, use data-driven
		bandwidth.
Output: selected bandwidth for the least-squares smoothed periodogram
        and Fisher test for white noise, Generalized likelihood test
	for white noise.  Further, the following results are reported:
         grid Points,LS-est-log(spectral),its bias,its SE,LLK-est-log(spectral) 
         After these five columns, the periodogram is appended
		Fundamental frequencies,     Periodogram

Source: This program was written to compute simulations and real data
        examples in the paper by Fan and Kreutzberger(1998)

Reference: Fan, J. and Gijbels, I. (1995a). Data-driven bandwidth
        selection in local polynomial fitting: variable bandwidth
        and spatial adaptation. {\em J. Royal Statist. Soc. B},
        {\bf 57}, 371--394.

        Fan, J. and Kreutzberger, E. (1998).  Automatic local smoothing for
        spectral density estimation.  {\em Scandi. Jour. Statist.},
        {\bf 25}, 359-369.

Last modified:  6-16-00

/* ***************** SIMULATED EXAMPLES  *********************************
  ARMA Models:    X_t + sum_{j=0}^{p-1} a_j X_{t-j-1} =
	     e_t + sum_{j=0}^{q-1} a_j e_{t-j-1}
  Parameters: a, b are given in each example.  See e.g. case 1.
  OUTPUT:    mhat[ii][0], mhat[ii][1], mhat[ii][2] contain LS estimated
             curve, bias and variance, respectively.  mhat[ii][3]
	     contains Local likelihood estimated curve.
For real data example, see above comments 
 ***********************************************************************  */

#include <stdio.h>
#include <math.h>
#define PI 3.14159
#define TWOSTAGE 1                           /* 1 means yes, otherwise no */
#define AORDER 2
#define NU     0
#define ADJUST (NU == 0? (TWOSTAGE == 1 ? .7776:.8941):(TWOSTAGE == 0 ? .7639:.7643) )
#define NORMIAL         1.96               /* confidence level */
#define MIN(i,j) ( (i) > (j) ? (j):(i) )
main()
{
   double *data,*x,*y,*ytrue,*xgrid,*z,**mhat,*coverage,*MISE,*MISEswap,*mtrue;
   double hopt,hgiven,xgridmin,xgridmax,  tmp, tmp1, tmp2, tmp3, sigma, *a, *b;
   double Imax, Imean, m0, lik, df;
   int n, ngrid, nsim, Nsim, i, j, k, order, idum, Idum[3];
   int kfixed, iexample, aorder;
   FILE *reg, *in;
   double *vector(), **matrix(),ran1(), gasdev(), maspec(), llk();
   void sort(), constband(), free_matrix(), free_vector(), exit();
   char filename[15], filename0[15];

          /* INITIALIZATION */

                          /*   order:   The order of local polynomial
                                        approximation
                          */
  if(NU == 0) printf("Curve Estimation ADJUST = %10.6f\n", ADJUST);
  else printf("Derivative Estimation ADJUST = %10.6f\n", ADJUST);
  if( TWOSTAGE == 1) printf("TWO-STAGE:   ADJUST = %10.6f\n", ADJUST);
  else   printf("ONE-STAGE:   ADJUST = %10.6f\n", ADJUST);
   order = 1+NU;
   ngrid = 201;
   printf("Which Example (0 -- 4) ?\n Example 0 is real data example:      ");
   scanf("%d", &iexample);
   if(iexample == 0)
   { printf("Enter sample size n (ngrid=201, xgridmin=0, xgridmax=PI): ");
     /*scanf("%d %d %lf %lf",&n, &ngrid, &xgridmin, &xgridmax);*/
     scanf("%d", &n); ngrid=201; xgridmin=0.0; xgridmax=PI;
  printf("Enter bandwidth: NEGATIVE VALUE means using DATA-DRIVEN bandwidth: ");
     scanf("%lf", &hgiven);
     printf("Enter input filename:  ");
     scanf("%s", filename0);
     printf("Enter OUTPUT filename: ");
     scanf("%s", filename);
     Nsim = 1;         /*dummy variable*/
   }
  else
   {
    printf("Enter n, Nsim:     \n");
    scanf("%d %d", &n, &Nsim);
    if(TWOSTAGE ==  1)
       sprintf(filename, ".reg%d.%d", iexample, n);
    else 
        sprintf(filename, ".regone%d.%d", iexample, n);
    reg       = fopen(filename, "w");
    hgiven = -1.0;
   }



  /* MEMORY ALLOCATION */
 xgrid=vector(0,ngrid);
 mtrue = vector(0, ngrid);
 a=vector(0,12);
 b=vector(0,6);
 x=vector(0,n);
 y=vector(0,n);
 z=vector(0,n);
 ytrue = vector(0,n);
 mhat=matrix(0,ngrid,0,3);
 coverage = vector(0,ngrid);            /* % of coverage probability */
 MISE = vector(0,Nsim);
 MISEswap = vector(0, Nsim);
            /*  INITIALIZATION , computing GRID POINTS */
 tmp = PI / (double) (ngrid -1);    
 for(i=1, xgrid[0]=0; i < ngrid; i++)
     { xgrid[i] = xgrid[i-1]+tmp;
       coverage[i] = 0.0;
     }
 tmp = PI / (double) n;       /* fixed point uniform design */
 for(i=1, x[0] = 0; i < n; i++) x[i] = x[i-1]+tmp;

    switch(iexample)
      {
      case 0: in = fopen(filename0, "r"); /*real data*/
            reg = fopen(filename,"w");
            for(i=0; i < n; i++)
                 fscanf(in, "%lf", z+i);

            /* Naive Implementation of Computing Periodogram */
	    tmp =  2*PI / (double) n; 
	    for(k=0; k < n/2; k++)
	     { tmp1 = (k+1)*tmp; 
	       tmp2 = 0.0; tmp3 = 0.0;
	       for(j=0; j < n; j++)
	          { tmp2 += z[j]* cos(tmp1*(j+1));
	            tmp3 += z[j]* sin(tmp1*(j+1));
                  }
               x[k] = tmp1;
	       y[k] = (tmp2*tmp2 + tmp3*tmp3)/n/2/PI;
	       y[k] = log(y[k]);
	     }

            for(i=0; i < ngrid; i++)  xgrid[i] = xgridmin+i*
                                 (xgridmax-xgridmin)/(ngrid -1.0);
            constband(x, y, n/2, order, xgrid, ngrid, mhat, &hopt,hgiven);
	    for(j=0; j < ngrid; j++) mhat[j][0] += 0.57721;
            /* Using local likelihood */
            for(j=0; j < ngrid; j++)
              mhat[j][3] = llk(x,y,n/2,hopt/1.1,xgrid[j],mhat[j][0]);

           /* ******* Compute Fisher Statisitc for white noise  ******** */

            Imax = 0.0; Imean = 0.0;  /* maximum and mean of the periodogram */
            for(k=0; k < n/2; k++)
              { tmp = exp(y[k]);
                Imean += tmp;
                if(tmp > Imax) Imax = tmp;
              }

            k=n/2;
            Imax = k*Imax/Imean;                         /* *Fisher statistic */
            Imean = 1 - exp(- exp( - Imax + log((double) k)));  /* P-value */

           /* ********* compute generalized likelihood statistic ******* */
            tmp1 = 0.0; tmp2 = 0.0;
            for(i=0; i < n; i++)
             { tmp1 += z[i];
               tmp2 += z[i]*z[i];
             }
            tmp3 = (tmp2 - tmp1*tmp1/n)/(n-1);   /* sample var of z's */
            m0 = log(tmp3 / 2 /PI);

            lik = 0.0;
            for(k=0; k < n/2; k++)
            {  tmp1 = x[k]*(ngrid-1.0)/PI;
               j    = tmp1;                     /* x[k] in jth interval */
               tmp2 = tmp1 - j;                 /* remainder  */
               tmp3 = tmp2*mhat[MIN(j+1,ngrid-1)][3] + (1-tmp2)*mhat[j][3];
                                                /* linear interpolation */
               lik += exp(y[k]-m0) - exp(y[k]-tmp3) + m0 - tmp3;
            }

            df = PI/(hopt/1.1)*0.45*2.1522;




         /* ************* Printing result statistic **************** */


      fprintf(reg, "Real data: ngrid=%d,xgridmin=%10.4f,xgridmax=%10.4f,NU=%d",
                       ngrid, xgridmin, xgridmax, NU);
            fprintf(reg, "\n estimated LS-bandwidth is %10.6f", hopt);
            fprintf(reg, "\n Periodogram is appended at the end of the file");
	    fprintf(reg, "\n Fisher statistic = %10.6f, P-value = %10.6f",
	    		Imax, Imean);
	    fprintf(reg, "\n r_K * Generalized likelihood statistic = %10.6f, ",
	    			2.1522*lik);
	    fprintf(reg, "degree of freedom = %10.5f", df);

            fprintf(reg, "\n GRID POINTS,   LS-Est,   Bias,   SE,   LLK-Est");
            fprintf(reg, "\n\n\n\n\n");
       for(j=0;j<ngrid; j++)fprintf(reg,"%12.5f %12.5f %12.5f %12.5f %12.5f\n",
                 xgrid[j], mhat[j][0], mhat[j][1], mhat[j][2],mhat[j][3]);
            fprintf(reg, "\n 2 pi k/ n,  periodogram");
            fprintf(reg, "\n\n\n\n\n");
       	    for(j=0;j<n/2;j++)fprintf(reg,"%12.5f %12.5f\n", x[j],y[j]);
            exit(1);

       case 1: a[0]=-1.5; a[1]=0.7; a[2]=-0.1;
               for(i=0; i < ngrid; i++)
                      mtrue[i]= log(1/maspec(xgrid[i],a,3)/2/PI);
               for(i=0; i < n; i++)
                    ytrue[i]=  log(1/maspec(x[i],a,3)/2/PI);
                break;

       case 2: b[0]=-0.3; b[1]=-0.6;b[2]=-0.3; b[3]=0.6;
               for(i=0; i < ngrid; i++)
                      mtrue[i] = log(maspec(xgrid[i],b,4)/2/PI);
                for(i=0; i < n; i++)
                      ytrue[i] = log(maspec(x[i],b,4)/2/PI);
                break;


       case 3:  a[0]=-1.5; a[1]=0.7; a[2]=-0.1;
		b[0]=-0.3; b[1]=-0.6;b[2]=-0.3; b[3]=0.6;
                for(i=0; i < ngrid; i++)
                   { mtrue[i]= 1/maspec(xgrid[i],a,3)+16*maspec(xgrid[i],b,4);
                     mtrue[i] = log(mtrue[i]/2/PI);
                   }
                for(i=0; i < n; i++)
                   { ytrue[i]= 1/maspec(x[i],a,3)+ 16*maspec(x[i],b,4);
                     ytrue[i] = log(ytrue[i]/2/PI);
                   }
               break;

      case 4: for(i=0; i < 12; i++) a[i] = 0.0;
	      a[3]=-0.9; a[7]=-0.7; a[11]=0.63;
      
              for(i=0; i < ngrid; i++)
                  mtrue[i]=log(1/maspec(xgrid[i],a,12)/2/PI); 
              for(i=0; i < n; i++)
                  ytrue[i]=log(1/maspec(x[i],a,12)/2/PI);
              break;                                  	       

     default:  printf("Out of Example list:\n");
               exit(1);
     }

fprintf(reg, "EXAMPLE %d \n", iexample);
fprintf(reg, "sample size n = %d, number of simulations = %d \n", n, Nsim);
fprintf(reg, "True log-spectrum:\n");
for(j=0; j < ngrid; j++) fprintf(reg,"%10.6f %10.6f\n", xgrid[j], mtrue[j]);
fprintf(reg, "THE ESTIMATED BANDWIDTH AND THE ESTIMATED CRUVES ARE: \n");
fprintf(reg, "(opt bandwidth and MADE first, then curve estimates and \n");
fprintf(reg, " their associated BIAS and VARIANCE at 90%, 50% and 10%)\n \n\n");
fprintf(reg, "opt band LS-MADE    LS-exp-MADE     LLK-MADE   LLK-exp-MADE ");
fprintf(reg, "LS-MASE    LS-exp-MASE     LLK-MASE   LLK-exp-MASE \n\n\n" );


/* *************************************************************
                    SIMULATION STARTS
   ************************************************************  */


  for(nsim=0; nsim < Nsim; nsim++)
    {
     idum = -10*nsim*nsim-2+nsim;
                      /* GENERATE DATA */
     for(i=0; i < n; i++) y[i] =  ytrue[i] +log(-log(ran1(&idum)));


    constband(x, y, n, order, xgrid, ngrid, mhat, &hopt,hgiven);
    fprintf(reg, "%12.6f", hopt); 
    for(j=0; j < ngrid; j++) mhat[j][0] += 0.57721;
            /* Using local likelihood */
    for(j=0; j < ngrid; j++) 
              mhat[j][3] = llk(x,y,n,hopt/1.1,xgrid[j],mhat[j][0]);


     /* compute MADE and the such */
    tmp = 0.0; tmp1=0.0; tmp2=0.2; tmp3=0.0;
    for(j=0; j < ngrid; j++) tmp +=  fabs(mhat[j][0] - mtrue[j]);  
    for(j=0; j < ngrid; j++) 
	    tmp1 += fabs(MIN(exp(mhat[j][0]),10*exp(mtrue[j]))-exp(mtrue[j]));  
    for(j=0; j < ngrid; j++) tmp2 += fabs(mhat[j][3] - mtrue[j]);
    for(j=0; j < ngrid; j++) 
	   tmp3 += fabs(MIN(exp(mhat[j][3]),10*exp(mtrue[j])) - exp(mtrue[j]));
    tmp /= ngrid; tmp1 /= ngrid;  tmp2 /= ngrid; tmp3 /= ngrid;
    MISE[nsim] = tmp2;
    fprintf(reg, "%12.6f %12.6f %12.6f %12.6f  ",tmp,tmp1,tmp2,tmp3);

    /*  compute MISE, weighted L2 instead of L1-loss  */

    tmp = 0.0; tmp1=0.0; tmp2=0.2; tmp3=0.0;
    for(j=0; j < ngrid; j++) 
	   tmp +=  (mhat[j][0] - mtrue[j]) * (mhat[j][0] - mtrue[j]);
    for(j=0; j < ngrid; j++) 
	    tmp1 += (MIN(exp(mhat[j][0]),10*exp(mtrue[j]))-exp(mtrue[j]))  
		    * (MIN(exp(mhat[j][0]),10*exp(mtrue[j]))-exp(mtrue[j]));
    for(j=0; j < ngrid; j++) 
	    tmp2 += (mhat[j][3] - mtrue[j])* (mhat[j][3] - mtrue[j]);
    for(j=0; j < ngrid; j++) 
	   tmp3 += (MIN(exp(mhat[j][3]),10*exp(mtrue[j])) - exp(mtrue[j]))
		   * (MIN(exp(mhat[j][3]),10*exp(mtrue[j])) - exp(mtrue[j]));
    tmp /= ngrid; tmp1 /= ngrid;  tmp2 /= ngrid; tmp3 /= ngrid;
    tmp = sqrt(tmp); tmp1 = sqrt(tmp1); tmp2 = sqrt(tmp2); tmp3=sqrt(tmp3);
    fprintf(reg, "%12.6f %12.6f %12.6f %12.6f\n",tmp,tmp1,tmp2,tmp3);

    printf("Example %d,  n = %d, nsim =%d, MISE=%10.6f, hopt =%10.6f\n",
	iexample, n, nsim, MISE[nsim], hopt);
  }     /*finish the loop with nsim */

/*  fprintf(reg, "Coverage Probability\n");
  for(j=0; j < ngrid; j++) fprintf(reg, "%7.4f \n", coverage[j]/Nsim); */

  fprintf(reg, "\n\n LS-Estimate   biases    variances   LLK-Est\n\n\n" );

  for(i=0; i < Nsim; i++) MISEswap[i] = MISE[i];
  sort(MISE, Nsim);
  i=0.1*Nsim;  tmp = MISE[i];  /*10th percentile's random seed */
  for(i=0; i < Nsim; i++) 
	if(fabs(MISEswap[i] - tmp) < 0.00001) Idum[0] = -10*i*i-2+i;
  i =0.5*Nsim;  tmp = MISE[i];  /*50th percentile's random seed */
  for(i=0; i < Nsim; i++)
        if(fabs(MISEswap[i] - tmp) < 0.00001) Idum[1] = -10*i*i-2+i;
  i=0.9*Nsim;  tmp = MISE[i];  /*90th percentile's random seed */
  for(i=0; i < Nsim; i++)
        if(fabs(MISEswap[i] - tmp) < 0.00001) Idum[2] = -10*i*i-2+i;
  printf("%d, %d, %d\n", Idum[0], Idum[1], Idum[2]);
      /* NOW RESIMULATE USING THE OLD SEED */
 for(nsim=0; nsim < 3; nsim++)
    {
    idum = Idum[nsim];
                      /* GENERATE DATA */
    for(i=0; i < n; i++) y[i] =  ytrue[i] +log(-log(ran1(&idum)));
    constband(x, y, n, order, xgrid, ngrid, mhat, &hopt,hgiven);
    for(j=0; j < ngrid; j++) mhat[j][0] += 0.57721;

    for(j=0; j < ngrid; j++)
		  mhat[j][3] = llk(x,y,n,hopt/1.1,xgrid[j],mhat[j][0]);
    for(j=0; j < ngrid; j++) fprintf(reg, "%12.5f %12.5f %12.5f %12.5f\n", 
	     mhat[j][0], mhat[j][1], mhat[j][2], mhat[j][3]);
    fprintf(reg, "\n \n \n");

    tmp = 0.0;
    for(j=0; j < ngrid; j++) tmp += fabs(mhat[j][3] - mtrue[j]);
    MISE[nsim] = tmp;
    tmp = 0.0;
    for(j=0; j < ngrid; j++) tmp += fabs(mhat[j][0] - mtrue[j]);
    MISE[nsim] = tmp;
    printf("Example %d,  n = %d, MISE=%10.6f, hopt =%10.6f\n",
        iexample, n, tmp/ngrid, hopt);
  }     /*finish the loop with nsim */


}

/* ********************* MASPECT.C ****************** 
   compute the function
     f(x) = | 1 + a_0 exp(-i x) + ... + a_{p-1} exp(-i p x) |^2
************************************************** */
double maspec(x,a,p)
double x, *a;
int p;
{int j;
 double sum1, sum2;
 sum1=1.0; sum2=0.0;
 for(j=0; j < p; j++)
   { sum1 += a[j]* cos((j+1.0)*x);
     sum2 += a[j]*sin((j+1.0)*x);
   }
 sum1 = sum1*sum1 + sum2*sum2;
 return sum1;
}

/* ************************** LLK.C  *********************** 
Compute the local log-likelihood for the log-linear model via
Fisher Scoring:
	 y = m(x) + epsilon where log(epsilon) ~ exp(1); namely
	 epsilon has a density exp(-exp(e))exp(e).
x, y:  n-vector, n is sample size, h is the bandwidth.
xx:  the point value has to be returned.
a0:  initial values
************************************************************ */

double llk(x,y,n,h,xx, a0)
double *x, *y, h, xx, a0;
int n;
#define Tiny 1.0e-12
{ int i, j, iter, nrow;
  double a, b, suma, sumb, a1, b1, tmp, *Xdes, *Y, *w, *vector(), a11,a12,a22;
  void free_vector();

  Xdes = vector(0,n);      
  Y    = vector(0,n);
  w    = vector(0, n);  
  suma=0.0; sumb=0.0; nrow=0; a1=a0, b1 = 0.0; nrow=0;
  a11=0.0; a12=0.0; a22=0.0;

/*   ***************** Computing local neighborhood ************* */
 for(i=0; i < n; i++)
      {tmp = (xx - x[i])/h; tmp *=tmp;
       if(tmp < 1.0)
          {w[nrow] = 0.75*(1.0 - tmp)/h;   /* Epanechnikov kernel */
           Xdes[nrow] = x[i] - xx;
           Y[nrow]    = y[i];
           tmp = exp( Y[nrow] - a1 - b1*Xdes[nrow]);
           suma += (tmp-1.0) * w[nrow];
           sumb += (tmp-1.0) * Xdes[nrow]*w[nrow];
	   a11  += tmp* w[nrow];
	   a12  += tmp*Xdes[nrow]*w[nrow];
	   a22  += tmp*Xdes[nrow]*Xdes[nrow]*w[nrow];
           nrow++;
          }
       }

 tmp =  a11*a22 - a12*a12;
 if(tmp < Tiny) goto lab3 ;
 a = a1 + (a22*suma - a12*sumb)/tmp;
 b = b1 + (a11*sumb - a12*suma)/tmp;

/* **************** Start iteration *********************** */
iter = 0;
while( fabs(a - a1) + fabs(b - b1) > 0.00001 && iter < 100 )
  { /* printf("a=%8.4f,a1=%8.4f,b=%8.4f, b1=%8.4f iter=%d\n",a,a1,b,b1,iter);*/
    a1 = a; b1 = b;
    suma = 0.0; sumb = 0.0; a11=0.0; a12=0.0; a22=0.0;
    for(i=0; i < nrow; i++)
      { tmp = exp( Y[i] - a1 - b1*Xdes[i]);
        suma += (tmp-1.0) * w[i];
        sumb += (tmp-1.0) * Xdes[i]*w[i];           
	a11  += tmp* w[i]; 
        a12  += tmp*Xdes[i]*w[i];
        a22  += tmp*Xdes[i]*Xdes[i]*w[i];
      }
    tmp =  a11*a22 - a12*a12;
    if(tmp < Tiny) goto lab3 ;
    a = a1 + (a22*suma - a12*sumb)/tmp;
    b = b1 + (a11*sumb - a12*suma)/tmp;
    iter++;
  } 
if(iter <= 1 || iter >= 99) a = a0;  /* use the initial value */
goto lab4;
lab3: a = a0;
lab4: free_vector(Xdes, 0,n); 
 free_vector(Y, 0,n);
 free_vector(w, 0, n);          
 return a;
}


/* ************************* CONSTBAND.C ******************************
   ******************************************************************
  Main program: Constant.C : select the bandwidth
  via a loop through bandwidth and do the fixed-order polynomial
  approximation.
  Calls for: LLS.C : which is the program doing the locally weighted
                         least squares regression

  Other programs called are:
               GAUSSJ.C : Solves a linear equation using Gauss-Jordan
                          elimination (see e.g. the book "Numerical
                          Recipes in C", by Press,Flannery, Teukolsky and
                          Vetterling, Cambridge University Press (1988))
               SORT.C : sorting a vector
               RAN1.C : random number generator for uniform (0,1)
                        distributed random variables
               GASDEV.C : random number generator, generates standard
                          normal distributed random variables; relies
                          on the Box-Muller method and uses the
                          random number generator for uniform random
                          variables (i.e calls RAN1.C)
               Utility Programs:
                  IVECTOR.C : memory allocation for integer-valued vector
                  VECTOR.C : memory allocation for vector
                  MATRIX.C : memory allocation for matrix

                  FREE_MATRIX.C : frees the memory allocated by MATRIC
                  FREE_IVECTOR.C : frees the memory allocated by IVECTOR
                  FREE_VECTOR.C : frees the memory allocated by VECTOR

                  NERROR.C : prints an error message.
**************************************************************************
************************************************************************** */


void constband(x, y, n, order, xgrid, ngrid, mhat, hopt,hgiven)
double *x, *y, *xgrid, **mhat, *hopt, hgiven;
int n,order, ngrid;
#define LARGE 1.0e17
#define SPAN  1.1                             /*maximum number of iteration*/
#define IUP   3                               /*checking conv of iteration*/ 
#define MIN(i,j) ( (i) > (j) ? (j):(i) )
#define MAX(i,j) ( (i) > (j) ? (i):(j) )
#define ABS(i)   (  (i) >= 0  ? (i):-(i))

 /*      OPTIMAL BANDWIDTH AND REGRESSION ESTIMATION
         Input : x :n-column vector containing the covariates;
                 y : n-column vector with the responses;
                 order : 0 = local constant approximation
                         1 = local linear approximation,
                         2 = local quadratic approximation,
                         3 = local cubic approximation,
                         etc. .....,
                 xgrid : ngrid-column vector with the grid points in
                         which the final regression estimator mhat is
                         calculated;
                 hgiven: if hgiven > 0, use this bandwidth; otherwise
		 	 use data driven one.
         Output : mhat  estimated regression curve, bias and var at grid points;
		  hopt:  the optimal bandwidth.
  */

{
  int i, ii, iup, ismooth, j, nrow, nm, a;
  double **beta, **betatmp, *SE, *RSC;
  double  h, hmin, hmax,hopt1,MASE, MSEnew, MSEold, xx, xmin, xmax, tmp, sigma;
  double hopt2, MASEopt, *vector(), **matrix();
  void free_vector(), free_matrix();
  void sort(), lls();

                                   /*MEMORY ALLOCATION */
  nm = order+8;
  beta=matrix(0,nm,0,1);  
  betatmp = matrix(0,nm, 0, ngrid);
  SE=vector(0,nm);        RSC = vector(0,nm);

  xmin=x[0]; xmax=x[0];
  for(i=0; i < n; i++)
       {if(x[i] < xmin) xmin = x[i];
	if(x[i] > xmax) xmax = x[i];
        }
  hmin =  (xmax - xmin)*(order+2.0)/n;   /* exp(-0.8*log((double) n)); */
  hmax = (xmax - xmin)/2.0;

  if(hgiven > 0) {hopt2 = hgiven; a = AORDER; goto lab5;}

  
  /*   a = AORDER;    approximation order*/

 if(TWOSTAGE == 1)                               /* TWO-STAGE PLUGIN */
   { a = AORDER; h = (2.0 +NU)* hmin;}
 else
   { a = 0;  h = hmin;}

  hopt1 = h; iup = 0; 
  MSEnew = LARGE; MASEopt = LARGE;
                    /*SEARCH FOR OPTIMAL BANDWIDTH FOR LOCAL order+a FIT */
  while(iup < IUP && h < hmax)
   { MASE = 0.0;
     for(ii = 0; ii < ngrid; ii++)   /* for each grid point */
       {xx = xgrid[ii];
				 /* nrow = # of effective data points */
        for(nrow = 0, i=0; i < n; i++) if(fabs(xx - x[i]) < h) nrow++;
        if(nrow > 2*(order+a)) /* NOT Nearly singular case */
            { lls(x,y,n,order+a,xx,h, beta, SE, &sigma, RSC, 1);
	      MASE    += RSC[0];
            }
        else{iup=0; goto lab1;}
       }
      MSEold = MSEnew;
      MSEnew = MASE;
      if(MSEnew > MSEold) iup++; else iup = 0;
      if(MSEnew < MASEopt) {MASEopt = MSEnew; hopt1 = h;}


lab1: h *= SPAN;
   }      /*end-loop through h*/
   hopt1 *=  ADJUST;   
   hopt2 = hopt1;

if( TWOSTAGE == 1)
{
  h = (1.0 + NU)*hmin;  hopt2 = h;
  iup = 0; MSEnew = LARGE; MASEopt = LARGE;

      /* ************* ESTIMATE CURVATIVE FIRST ********** */
  for(ii =0; ii < ngrid; ii++) 
    {  xx = xgrid[ii];
       lls(x,y,n,order+a,xx,hopt1, beta, SE, &sigma, RSC, 0); /*(order+a)-fit */
       for(i=0; i <= order+a; i++) betatmp[i][ii] = beta[i][0];
     }
           /*END priliminary fit to learn beta_2 so that it can be passed  */

  while(iup < IUP && h < hmax)
   {MASE=0.0; 
    for(ii=0; ii < ngrid; ii++)
     { xx = xgrid[ii];
       for(i=0; i <= order+a; i++) beta[i][1] = betatmp[i][ii];
           /*END priliminary fit to learn beta_2 so that it can be passed 
	     to compute the bias NOW TRY LOCAL order-FIT,
	     nrow = # of effective data points */
       for(nrow = 0, i=0; i < n; i++) if(fabs(xx - x[i]) < h) nrow++;
       if(nrow <= order+a)                   /* Nearly singular case */
          {iup=0; goto lab2;}
       else
           { lls(x,y,n,order,xx,h, beta, SE, &sigma, RSC, a);
	      MASE += beta[NU][1]*beta[NU][1] + SE[NU]*SE[NU];
            }  
      }               /*end loop through ii */

    MSEold = MSEnew;
    MSEnew = MASE;
    if(MSEnew > MSEold) iup++; else iup = 0;
    if(MSEnew < MASEopt) {MASEopt = MSEnew; hopt2 = h;}

lab2: h *= SPAN;
   }                             /*end loop through h */
     /* hopt2 = h * exp(-(IUP+1)*log(SPAN)); */
           /* optimal bandwidth for local order-fit*/
 }                /* end-if with TWOSTAGE == 1 */
                /* NOW COMPUTE COMPUTE THE ESTIMATED CURVE */

lab5: for(ii=0; ii < ngrid; ii++)  
   { xx = xgrid[ii];
     lls(x,y,n,order,xx,hopt2, beta, SE, &sigma, RSC, a);
     mhat[ii][0] = beta[NU][0];
     mhat[ii][1] = beta[NU][1];
     mhat[ii][2] = SE[NU];
   }                               /*end the loop through ii */

   hopt[0] = hopt2;
    free_matrix(beta,0,nm, 0, 1); 
    free_matrix(betatmp,0,nm,0,ngrid);
    free_vector(SE,0,nm); 
    free_vector(RSC, 0, nm);
}


/* *************  GAUSSJ.C ************** */
/* #include <math.h>
#include <stdio.h>  */
#define SWAP(a,b) {double temp=(a);(a)=(b);(b)=temp;}
#define TINY 1.0e-12;
void gaussj(a,n,b,m)
     /* Linear equation solution by Gauss-Jordan elimination
        Input: a: n by n input matrix
               b: n by m input matrix containing the m right-hand sides
                  vectors of the equation-system
        In the present version, b is a column vector (m=1). The program
        has to be changed slightly for the general case (m > 1). In this
        case, b should be declared as a matrix and e.g. line 58 should
        look like:
            for (l=1; l<=m; l++) SWAP(b[irow-1][l-1], b[icol-1][l-1])
        Output: a: the input matrix a is replaced be its matrix inverse
                b: the input matrix is replaced by the corresponding set
                   of solution vectors. */

double **a,**b;
int n,m;

{
  int *indxc,*indxr,*ipiv;
  int i,icol,irow,j,k,l,ll, *ivector();
  double big,dum,pivinv;
  void nrerror(), free_ivector();

  indxc=ivector(1,n);
  indxr=ivector(1,n);
  ipiv=ivector(1,n);


  for (j=1; j <= n; j++) ipiv[j]=0;
  for (i=1; i <= n; i++) {
        big=0.0;
        for (j=1; j <= n; j++)
            if (ipiv[j] !=1)
                 for (k=1; k<=n; k++) {
                   if (ipiv[k] == 0) {
                       if (fabs(a[j-1][k-1]) >= big) {
                           big=fabs(a[j-1][k-1]);
                           irow=j;
                           icol=k;
                       }
                   } else if (ipiv[k] > 1) nrerror("GAUSSJ: Singular Matrix-1");
            }
        ++(ipiv[icol]);

        if (irow !=icol) {
            for (l=1; l<=n; l++) SWAP(a[irow-1][l-1], a[icol-1][l-1])
            for (l=1; l<=m; l++) SWAP(b[irow-1][l-1], b[icol-1][l-1])
        }
        indxr[i]=irow;
        indxc[i]=icol;
        if (fabs(a[icol-1][icol-1]) == 0.0) a[icol-1][icol-1] = TINY;
        pivinv=1.0/a[icol-1][icol-1];
        a[icol-1][icol-1]=1.0;
        for(l=1;l<=n;l++) a[icol-1][l-1] *=pivinv;
        for(l=1;l<=m;l++) b[icol-1][l-1] *=pivinv;
        for (ll=1;ll<=n;ll++)
            if (ll !=icol) {
                dum=a[ll-1][icol-1];
                a[ll-1][icol-1]=0.0;
                for (l=1;l<=n;l++) a[ll-1][l-1] -= a[icol-1][l-1]*dum;
                for (l=1;l<=m;l++) b[ll-1][l-1] -= b[icol-1][l-1]*dum;
            }
  }
  for (l=n; l>=1;l--) {
      if (indxr[l] !=indxc[l])
        for(k=1; k <=n; k++)
                      SWAP(a[k-1][indxr[l]-1],a[k-1][indxc[l]-1]);
  }
  free_ivector(ipiv,1,n);
  free_ivector(indxr,1,n);
  free_ivector(indxc,1,n);
}

/*       ********************* SORT.C ***************** */
void sort(ra,n)
int n;
double *ra;
/*index of ra is 0:n-1*/
{
        int l,j,ir,i;
        double rra;

        l=(n >> 1)+1;
        ir=n;
        for (;;) {
                if (l > 1)
                        rra=ra[--l-1];
                else {
                        rra=ra[ir-1];
                        ra[ir-1]=ra[0];
                        if (--ir == 1) {
                                ra[0]=rra;
                                return;
                        }
                }
                i=l;
                j=l << 1;
                while (j <= ir) {
                        if (j < ir && ra[j-1] < ra[j]) ++j;
                        if (rra < ra[j-1]) {
                                ra[i-1]=ra[j-1];
                                j += (i=j);
                        }
                        else j=ir+1;
                }
                ra[i-1]=rra;
        }
}

/* ******************** UNIFORM RANDOM NUMBER GENERATOR: RAN1.c ******* */
#include <stdlib.h>
#define M1 259200
#define IA1 7141
#define IC1 54773
#define RM1 (1.0/M1)
#define M2 134456
#define IA2 8121
#define IC2 28411
#define RM2 (1.0/M2)
#define M3 243000
#define IA3 4561
#define IC3 51349

double ran1(idum)
int *idum;

                    /* returns a uniform random deviate between 0.0 and 1.0.
                       Set idum to any negative value to initialize or
                       reinitialize the sequence  */

{
    static long ix1,ix2,ix3;
    static double r[98];
    double temp;
    static int iff=0;
    int j;
    void nrerror();

    if (*idum < 0 || iff == 0) {
         iff=1;
         ix1=(IC1-(*idum)) % M1;
         ix1=(IA1*ix1+IC1) % M1;
         ix2=ix1 % M2;
         ix1=(IA1*ix1+IC1) % M1;
         ix3=ix1 % M3;
         for (j=1;j<=97;j++) {
               ix1=(IA1*ix1+IC1) % M1;
               ix2=(IA2*ix2+IC2) % M2;
               r[j]=(ix1+ix2*RM2)*RM1;
           }
           *idum=1;
    }
    ix1=(IA1*ix1+IC1) % M1;
    ix2=(IA2*ix2+IC2) % M2;
    ix3=(IA3*ix3+IC3) % M3;
    j=1 + ((97*ix3)/M3);
    if (j > 97 || j < 1) nrerror("RAN1: This cannot happen.");
    temp=r[j];
    r[j]=(ix1+ix2*RM2)*RM1;
    return temp;
}

/* ***************** NORMAL RANDOM GENERATOR: GASDEV.C *********** */
double gasdev(idum)
int *idum;
                   /* returns a normally distributed deviate with mean 0
                      and variance 1, using ran1(idum) as the source of uniform
                      deviates    */

{
    static int iset=0;
    static double gset;
    double fac,r,v1,v2;
    double ran1();

    if (iset == 0)  {
        do {
            v1=2.0*ran1(idum)-1.0;
            v2=2.0*ran1(idum)-1.0;
            r=v1*v1+v2*v2;
        } while (r >= 1.0 || r == 0.0);
        fac= sqrt(-2.0*log(r)/r);
        gset=v1*fac;
        iset=1;
        return v2*fac;
    }  else {
         iset=0;
         return gset;
    }
}

/*                    UTILITY PROGRAMS  */
/* ************************************************************************* */
/*        *************** IVECTOR.C  **********    */
int *ivector(nl, nh)
int nl, nh;
{  int *v;
   v = (int *)malloc((unsigned) (nh-nl+1) * sizeof(int));
   return v-nl;
}

/*        *************** VECTOR.C  **********    */
double *vector(nl, nh)
int nl, nh;
{  double *v;
   v = (double *)malloc((unsigned) (nh-nl+1) * sizeof(double));
   return v-nl;
}


/*        ***************MATRIX.C  **********    */
double **matrix(nrl,nrh,ncl,nch)
int nrl,nrh,ncl,nch;

        /* Allocates a double matrix with range [nrl..nrh][ncl..nch]    */

{
    int i;
    double **m;

    m=(double **) malloc((unsigned) (nrh-nrl+1)*sizeof(double*));
    m -=nrl;

    for(i=nrl;i<=nrh;i++) {
         m[i]=(double *) malloc((unsigned) (nch-ncl+1)*sizeof(double));
         m[i] -= ncl;
    }
    return m;
}

/*        ***************FREE_MATRIX.C  **********    */
void free_matrix(m,nrl,nrh,ncl,nch)
double **m;
int nrl,nrh,ncl,nch;
                     /* Frees a matrix allocated with matrix */

{
    int i;

    for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
    free((char*) (m+nrl));
}

/*        ***************FREE_IVECTOR.C  **********    */
void free_ivector(v,nl,nh)
int *v,nl,nh;
                       /* Frees an int vector allocated by ivector() */

{
     free((char*) (v+nl));
}

/*        ***************FREE_VECTOR.C  **********    */
void free_vector(v,nl,nh)
double *v;
int nl,nh;
                       /* Frees a double vector allocated by vector() */

{
     free((char*) (v+nl));
}


/*    *********************NRERROR.C ******************** */
void nrerror(error_text)
     /* prints an error message */

char error_text[];
{
   fprintf(stderr,"Numerical recipes run-time error...\n");
   fprintf(stderr,"%s\n",error_text);
   fprintf(stderr,"...now exiting to system...\n");
   exit(1);
}

/* *******************************LLS.C ************************ */
void lls(x,y,n,p,xx,h, beta, SE, sigma, RSC, a)
double *x, *y,xx,h, **beta, *SE,*sigma, *RSC;
int     n,  p, a;
  /*Do locally weighted least-squared by using local polynomial of order p
    Input: x is n-vector of design
	   y is n-vector of the response
	   xx is the point where the estimated value to be returned
	   h  is the bandwidth and the kernel is Epnechnikov;
           n and p explained above
	   beta: (p+a) by 2 matrix.
           beta[.][1] estimated coefficients to be used to estimate
		      bias: (p+a)-vector;
	   a: approximation order in the plub-in procedure and
              > 0, return regression coefficients along with others
	      = 0, return only estimated regression coefficient.

    Output:  beta[.][0]: estimated regression coefficients (0 -- p)
             beta[.][1]: estimated bias according to (2.7) (0 -- p).  
             SE:  p-column vector of estimated SE-values of
                  the estimated regression coefficients for the
                  first response vector.
                  t = beta/SE
             sigma: the estimated standard error of linear model.
             RSC:  Extrended cross-validated values computed as follows. 
  */
{
 int i,j,l,nrow, nmax;
 double tmp, tmp1, *Xdes,*Y, **H, **Ha, *w ;
 double RSS, sum, sumw, trace;
 double **matrix(), *vector(), *s, *S;
 void   free_matrix(), free_vector();
  
      /*MEMORY ALLOCATION*/
  nmax = MAX(p+a, 2*p+2);             /* max moments to be computed */
  Xdes = vector(0,n);                   Y    = vector(0,n); 
  H=matrix(0,p,0,p);                    Ha   = matrix(0,p,0,p);
  w    = vector(0, n);                  s    = vector(0,nmax);
  S    = vector(0, nmax);
 
      /* DETERMINING THE LOCAL DATA POINTS */
  nrow=0;                                 /* Effective number of data points*/
  sumw =0.0;                              /* Total weight */
  for(i=0; i < n; i++)
      {tmp = (xx - x[i])/h; tmp *=tmp;
       if(tmp < 1.0)
          {w[nrow] = 1.0 - tmp;                   /* Epanechnikov kernel */
           sumw += w[nrow];
           Xdes[nrow] = x[i] - xx;
           Y[nrow]    = y[i];
           nrow++;
          }
       }

      /* COMPUTE  s[j] = \sum w[i] (X[i]-x)^j,
                  S[j] = \sum w[i]^2 (X[i]-x)^j, beta[j][0] = X'Wy[j]  */

    for(j=0; j <= nmax; j++) {s[j]=0.0; S[j]=0.0;}
    for(j=0; j <= p; j++)      beta[j][0] =0.0;
    for(i=0; i < nrow; i++)
      { tmp = 1.0;                    /*tmp = Xdes[i]^j */
        for(j=0; j <= nmax; j++)
            {tmp1 = w[i]*tmp;         
             s[j] += tmp1;
             S[j] += w[i]*tmp1;   
	     if(j <= p) beta[j][0] += tmp1*Y[i];
             tmp *= Xdes[i];
            }
       }

   /*computing H= X'WX, Ha = X'W^2X*/
   for(i=0; i <=p; i++)
        for(j=0; j <=p; j++)
        	  {  H[i][j] = s[i+j]; Ha[i][j] = S[i+j]; }


			  /* compute beta[.][1] according to (2.7) */
  if(a > 0)
   {
     for(i=0; i <= p; i++) 
       { tmp = 0.0;
         for(j=p+1; j <= p+a; j++) 
	    if(j+i <= p+a) tmp += beta[j][1] * s[j+i];
         beta[i][1] = tmp; 
       }
    }


  gaussj(H,p+1,beta,2);            /* solve H b = beta;
                                     output: H <- H^{-1}, beta <- H^{-1} beta */

if(a > 0)
 {
  for(RSS=.0, i=0; i < nrow; i++)   /* compute weighted residual sum of squares
                                     RSS = sum_1^n (y_i - \hat{y}_i)^2 w_i
                                     for the 0th column of Y */
   {
     sum=0.0; tmp=1.0;               /* tmp = Xdes[i]^l */
     for(l=0; l <= p; l++)   
	   { sum += tmp * beta[l][0] ; tmp *= Xdes[i];}
     RSS += (Y[i] - sum) * (Y[i] - sum)*w[i];
   }


  trace = 0.0;                   /*trace =  tr((X'WX)^{-1} X'W^2X )*/
  for(i=0; i <= p ; i++)
       {  tmp = 0.0;
          for(j=0; j<i; j++) tmp += H[i][j] * Ha[j][i];
          trace += H[i][i]*Ha[i][i] + 2.0 * tmp;
       }
  trace = sumw - trace;          /*trace = tr(W) -  tr((X'WX)^{-1} X'W^2X )*/


  RSS =  RSS / trace;            /* standardarized residual sum of squares */



  for (l = 0; l <= p; l++)  
   { sum = 0.0;              /*lth element of (X'WX)^{-1}(X'W^2X)(X'WX)^{-1}*/
     for(i=0; i <= p; i++)
        {
	  tmp = 0.0;
          for(j=0; j<i; j++) tmp += Ha[i][j] * H[j][l];
          sum += H[i][l]*(Ha[i][i]*H[i][l]+2.0 * tmp);
        }
     SE[l] = sqrt(RSS *sum);
     RSC[l] = sum;
   }

  RSC[0] = RSS * (1.0 + (p+1)*RSC[0]);

  sigma[0] = sqrt(RSS);
  }  /* end-if */

  free_vector(Xdes, 0,n);               free_vector(Y, 0,n);
  free_matrix(H,0,p,0,p);               free_matrix(Ha,0,p,0,p);
  free_vector(w, 0, n);                 free_vector(s, 0,nmax);
  free_vector(S,  0, nmax);

}
